@@ -5,16 +5,44 @@ import math
 import networkx as nx
 from nltk.corpus import words as wn_words
 import statistics as stat
-from pprint import pprint
+from collections import OrderedDict
+#from pprint import pprint
+import json
 
 from .nlp import *
 from .sim import *
-from .refiner import refine
+from .refiner import refine, ask_bert
 from .vis import pshow,gshow
 
 client = NLPclient()
 
-words = set(wn_words.words())
+def my_path() :
+  '''
+  detects directory where package sources get installed
+  useful to load .json, .txt etc. resources from there
+  '''
+  if __name__=='__main__' :
+    return "./"
+  else :
+    return __file__
+
+def get_freqs() :
+  '''
+  loads .json file contining frequencies
+  for around 20k English words
+  '''
+  global lemma_freqs
+  if lemma_freqs : return lemma_freqs
+  fname=my_path().replace('talk.py','lemmas.json')
+  with open(fname,'r') as f:
+    lemma_freqs=json.load(f)
+    return lemma_freqs
+
+# set of WordNet words
+wnet_words = set(wn_words.words())
+
+# global holder for lemma frequences, filled out on first use
+lemma_freqs = None
 
 def run_with(fname,query=True) :
   '''
@@ -47,7 +75,7 @@ def chat_about(fname,qs=None) :
 
 
 def tprint(*args) :
-  ''' custom print when trance on'''
+  ''' custom print when trace on'''
   if trace : print(*args)
 
 
@@ -117,9 +145,14 @@ def digest(text) :
     sent_data.append(d)
   return sent_data,l2occ
 
+# names of components of fileds in json array
+# collecting results of corenlp, for faster processing
 SENT,LEMMA,TAG,NER,DEP,IE=0,1,2,3,4,5
 
 def to_sents (text) :
+  '''
+  generator of sentences from text string
+  '''
   sent_data,_= digest(text)
   for x in sent_data :
     yield x[SENT]
@@ -241,6 +274,9 @@ def rank_sort(pr) :
   ''' sort dict by ranks associatied to its keys'''
   by_rank=[(x,r) for (x,r) in pr.items()]
   by_rank.sort(key=lambda x : x[1],reverse=True)
+  #for x in take(100, by_rank):
+    #if not isinstance(x, int):
+      #ppp('BY RANK', x)
   return by_rank
 
 def ners_from(d):
@@ -277,7 +313,7 @@ def wn_from(l2occ) :
         yield (s, v, o)
 
 def v2rel(v) :
-  '''rewrites "be" lemma to is'''
+  '''rewrites "be" lemma to "i"s, for more natural reading of relations'''
   if v=='be' : return 'is'
   return v
 
@@ -352,23 +388,32 @@ def answer_quest(q,talker) :
     #ppp(i,b)
     rank, id, shared, sent = b
     answers.append((id, sent, round(rank, 4), shared))
-  if not talker.params.answers_by_rank: answers.sort()
-  #ppp('#####',answers)
+
   if talker.params.with_refiner:
     wss =  [ws for (_,ws,_,_) in answers]
-    wss=refine_wss(wss)
+    wss=refine_wss(wss,talker)
     answers=[(0,ws,0,set()) for ws in wss]
-
   return answers, answerer
 
-def refine_wss(wss):
+
+def refine_wss(wss,talker):
+    '''
+    refines a few dozen lists of lists of words (extractive summary or
+    set of answer senteces) with BERT, for a second opinion on
+    what the best shorter summary or answer would be
+    '''
     sents = []
     for ws in wss:
       sents.append(nice(ws))
+    if not wss: return wss
     input = " ".join(sents)
-    output = refine(input)  # <====== calling refiner
+    how=talker.params.with_refiner
+    output = refine(input,how)  # <====== calling refiner
     xss = list(to_sents(output))
-    return xss
+    wss=list(take(talker.params.top_answers,wss))
+    return [['DOCTALK:']]+wss+\
+           [['BERT REFINING LARGER TEXT EXTRACTED BY DOCTALK:']]+\
+           xss
 
 def sigmoid(x): return 1 / (1 + math.exp(-x))
 
@@ -379,7 +424,7 @@ def answer_rank(id,shared,sent,talker,expanded=0) :
   if not lshared : return 0
 
   sent_count=len(talker.db[0])
-  word_count=len(talker.db[1])
+  #word_count=len(talker.db[1])
 
   lsent = len(sent)
   lavg=talker.avg_len
@@ -431,14 +476,18 @@ def interact(q,talker):
   talker.say(q)
   print('')
   ### answer is computed here ###
-  answers,_=answer_quest(q, talker)
+  answers,answerer=answer_quest(q, talker)
   show_answers(talker,answers)
-  talker.distill(q)
+  talker.distill(q,answers,answerer)
 
 def show_answers(talker,answers) :
   ''' prints out/says answers'''
   print('ANSWERS:\n')
-  for info, sent, rank, shared in take(talker.params.top_answers,answers):
+  if not talker.params.with_refiner :
+    answers=take(talker.params.top_answers,answers)
+    if not talker.params.answers_by_rank:
+      answers=sorted(answers)
+  for info, sent, rank, shared in answers:
     if not talker.params.with_refiner :
        print(info,end=': ')
     talker.say(nice(sent))
@@ -451,9 +500,14 @@ class Talker :
   '''
   class aggregating summary, keyphrase, relation extraction
   as well as query answering in the form of extracted sentences
-  based on given file or text
+  based on given file or text or preprocessed json equivalent
   '''
-  def __init__(self,from_file=None,from_text=None,params=talk_params()):
+  def __init__(self,
+               from_json=None,
+               from_file=None,
+               from_text=None,
+               params=talk_params()
+               ):
     '''creates data container from file or text document'''
     self.params=params
 
@@ -463,8 +517,13 @@ class Talker :
        self.from_file=from_file
     elif from_text :
        self.db=digest(from_text)
+    elif from_json :
+      xs=json.loads(from_json)
+      assert isinstance(xs,list) and len(xs)==1
+      text=xs[0]
+      self.db = digest(text)
     else :
-      assert from_file or from_text
+      assert from_file or from_text or from_json
 
     self.avg_len = get_avg_len(self.db)
 
@@ -472,10 +531,64 @@ class Talker :
     self.svo_graph=None
 
     self.g,self.pr=self.to_graph()
-    #self.get_sum_and_words(sk,wk)
+
     self.summary, self.keywords = \
       self.extract_content(self.params.max_sum, self.params.max_keys)
+    assert self.by_rank != None
 
+  def get_summary(self):
+    '''
+    function  extracting highest ranked sentences as summary
+    '''
+    yield from take(self.params.top_sum,self.summary)
+
+  def get_keys(self):
+    '''
+       function for extracting highest ranked keywords
+     '''
+    yield from take(self.params.top_keys,nice_keys(self.keywords))
+
+  def summary_sentences(self):
+    '''
+      API function  extracting highest ranked sentences as summary
+      encoded as a list of list of words in json form
+     '''
+
+    wss=[x[2] for x in self.get_summary()]
+    return json.dumps(wss)
+
+  def keyphrases(self):
+    ''' API  function for extracting highest ranked keywords as
+        a json encoded list of words
+    '''
+    ks=json.dumps(list(self.get_keys()))
+    return ks
+
+  def answer_question(self,quest,is_json=False):
+    '''
+    answers question given as a string,
+    returns answer possibly in jsno form
+    '''
+    assert isinstance(quest, str)
+    if is_json :
+      qs = json.loads(quest)
+      q=qs[0]
+    else :
+      q =quest
+    xs,_=self.answer_quest(q)
+    rss=[a[1] for a in xs]
+    b=None
+    if self.params.with_bert_qa > 0 :
+      wss=[' '.join(ws) for ws in rss]
+      ws = ' '.join(wss)
+      b=ask_bert(ws,q,confid=self.params.with_bert_qa)
+    answers = list(take(self.params.top_answers,rss))
+
+    if b :
+       b="Short answer : "+b+" !"
+       bs=b.split(' ')
+       answers=[bs]+answers
+    return json.dumps(answers)
 
   def answer_quest(self,q):
     '''answers question q'''
@@ -505,6 +618,10 @@ class Talker :
     return self.db[1].get(lemma)
 
   def to_ids(self,nodes) :
+    '''
+    returns sentence ids for lemma nodes in the graph
+    by looking them up in the word-to-sentence occurrence map
+    '''
     ids=set()
     for w in nodes :
       occs=self.get_occs(w)
@@ -516,12 +633,19 @@ class Talker :
 
 
   def adjust_sent_ranks(self,pr):
+    '''
+    adjust sentence and jeyword ranks via heuristics
+    in normalize_sent and normalize_key
+    '''
     npr = dict()
     for x, r in pr.items():
       if isinstance(x, int):
         ws = self.db[0][x][SENT]
         r = normalize_sent(r, len(ws), self.avg_len)
+      else:
+        r = self.normalize_key(x,r)
       npr[x] = r
+
     return npr
 
   def get_sentence(self,i):
@@ -556,6 +680,10 @@ class Talker :
         return x
 
     def nice_word(x,good_tags='N',lift=False) :
+      '''
+        heuristics for filtering and prioritizing
+        words and compound words
+      '''
       ws_ts=self.get_tagged(x)
       if not ws_ts : return None
       ws, tags = ws_ts
@@ -565,6 +693,7 @@ class Talker :
           ncount += 1
       if ncount > len(tags) / 2:
         ns=self.g.adj.get(x)
+        # lifts compounds to higher ranks if that makes sense
         if lift and ns and not isinstance(x,tuple):
           min_rank=self.pr[x] / self.params.prioritize_compounds
           xss=[(n,self.pr[n])
@@ -576,66 +705,69 @@ class Talker :
               ]
           if xss:
             xss.sort(key=lambda v: v[1], reverse=True)
+            #ppp(xss)
             xs=xss[0][0]
             return xs
         return x
       else:
         return None
 
-    sents,words=list(),set()
+    sents,words=list(),OrderedDict()
     npr=self.adjust_sent_ranks(self.pr)
+    # ordering all by rank here
     by_rank=rank_sort(npr)
+    self.by_rank = by_rank
 
+    # collect best by rank, but adjusting some
     for i  in range(len(by_rank)):
       x,r=by_rank[i]
       if sk and isinstance(x,int) :
         ws=self.db[0][x][SENT]
         ls=self.db[0][x][LEMMA]
-        if not is_clean_sent(ls) : continue
+        if not is_clean_sent(ls,self.params.known_ratio) : continue
         sk-=1
         sents.append((r,x,ws))
       elif wk and good_word(x) :
-        x=nice_word(x,lift=True)
+        #ppp('PWS', x)
+        x=nice_word(x,lift=self.params.prioritize_compounds>0)
         if x:
           wk -= 1
-          words.add(x)
+          #ppp('LWS', x)
+          words[x]=wk
       elif wk and isinstance(x,tuple) :
-          xs=tuple(map(nice_word,x))
-          if all(xs) :
+          x=tuple(map(nice_word,x))
+          if all(x) :
             wk -= 1
-            words.add(xs)
+
+            words[x]=wk
+
+    # ordering sentences by id, not rank here
     sents.sort(key=lambda x: x[1])
-    summary=[(r,x,ws) for (r,x,ws) in sents]
-    self.by_rank=by_rank # to be used when needed
+    #for sss in sents : ppp(sss)
+    summary=sents
 
     # remove word if in a tuple that is also selected
     for xs in words.copy() :
       if isinstance(xs,tuple) :
         for w in xs:
           if w in words:
-            words.remove(w)
-    clean_words=set()
+            del words[w]
+    clean_words=OrderedDict()
     for xs in words :
       if isinstance(xs,tuple) :
-        clean_words.add(tuple(map(maybe_cap,xs)))
+        clean_words[tuple(map(maybe_cap,xs))]=True
       else :
-        clean_words.add(maybe_cap(xs))
+        clean_words[maybe_cap(xs)]=True
 
     if self.params.with_refiner:
-      sents = []
+      # mimics usual return types after refining with BERT
+      summary=sorted(summary,reverse=True,key=lambda x : x[0])
       wss=[ws for (_,_,ws) in summary]
-      wss=refine_wss(wss)
+      wss=refine_wss(wss,self)
       xss = [(0,0,ws) for ws in wss]
       summary = xss
 
-    return summary,clean_words
-
-    def distill(self,q) :
-      '''
-      superclasses can do here additional
-      processing of question q
-      '''
-      pass
+    return summary,list(clean_words)
 
   def to_svos(self):
     '''
@@ -675,6 +807,10 @@ class Talker :
     return d
 
   def to_word_orbit(self,lemma):
+    '''
+    extracts orbit of given lemma through sentence space
+    as  (sentence,rank) pairs
+    '''
     _, l2occ = self.db
     occs=l2occ.get(lemma)
     if not occs : return None
@@ -682,6 +818,10 @@ class Talker :
     return sranks
 
   def to_sent_orbit(self,id):
+    '''
+    extracts orbit of given sentence
+    through lemma  space as (word,rank) pairs
+    '''
     def pr_of(x) :
       r=self.pr.get(x)
       if r : return r
@@ -703,6 +843,39 @@ class Talker :
       g.add_edge(s,o,rel=v,occs=occs)
     return g
 
+
+  def to_dep_tree(self):
+    '''
+    extracts dependency graph (mostly a tree)
+    by fusing dependency trees of all sentences
+    '''
+    g=nx.DiGraph()
+    for f,r,t in self.dep_edge() :
+      g.add_edge(f,t,rel=r)
+      #print(f, r, t)
+    return g
+
+  def dep_edge(self):
+    ''' dependency endge generator'''
+    sent_data, l2occ = self.db
+    for info in sent_data:
+      ws,ls,ts,_,deps,_=info
+      for dep in deps:
+        #print(dep)
+        f, r, t = dep
+        if t== -1 : #  and r=='ROOT' :
+          wt='SENT'
+          tt = 'TOP'
+        else :
+          wt = ls[t]
+          tt=ts[t]
+        tf=ts[f]
+        wf=ls[f]
+        #r=r.replace(':','*')
+        trt="_".join([tt,r,tf])
+        yield wt, trt, wf
+
+
   def to_edges_in(self,id,sd):
     '''yields edges from dependency structure of sentence id'''
     for dep in dep_from(id, sd):
@@ -739,8 +912,7 @@ class Talker :
                 yield occ[0],lemma
 
   def to_graph(self, personalization=None):
-    ''' builds document graph from several link types'''
-    db=self.db
+    ''' builds document graph from several link types '''
     svos=self.svos
     g = nx.DiGraph()
     for e in self.to_edges():
@@ -754,8 +926,11 @@ class Talker :
         else:
           g.add_edge(o, s)
 
-      pr = nx.pagerank(g, personalization=personalization)
-      if self.params.use_line_graph and g.number_of_edges()<20000 :
+    if personalization == None and self.params.pers_idf :
+      personalization=self.pers_from_freq(get_freqs())
+
+    pr = nx.pagerank(g, personalization=personalization)
+    if self.params.use_line_graph and g.number_of_edges()<20000 :
         lg=nx.line_graph(g)
         lpr= nx.pagerank(lg)
         for xy,r in lpr.items() :
@@ -763,9 +938,32 @@ class Talker :
           if isinstance(x,str) and isinstance(y,str):
             pr[x]=pr[x]+r
             pr[y]=pr[y]+r
-
     return g, pr
 
+  def pers_from_freq(self,freqs):
+    ''' returns personalization dictionary derived from word frequencies'''
+    d=dict()
+    _,l2occ=self.db
+    for w,r in freqs.items() :
+      if w in l2occ and r>0:
+        p=(1+math.log(len(l2occ[w])))/math.log(1+r)
+        d[w]=p
+    return d
+
+  def normalize_key(self,w,r):
+    ''' heuristics for normalizing keyphrase ranks'''
+    if not self.params.use_freqs : return r
+    freqs=get_freqs()
+    _, l2occ = self.db
+    if not w in freqs or not w in l2occ : return r
+    fr=freqs[w]
+    approx_tf_idf=(1+math.log(len(l2occ[w]))) / math.log(1 + fr)
+    p=(2*r*approx_tf_idf)/(r+approx_tf_idf)
+    #p=math.sqrt(p)
+    #p = r * approx_tf_idf
+    return p
+
+
   def to_prolog(self):
     ''' generates a Prolog representation of a document's content'''
     if not self.from_file : return
@@ -786,8 +984,33 @@ class Talker :
         occs=sorted(occs)
         f.write(f'svo{s,v,o,occs}.\n')
 
-  def distill(self, q):
-    pass
+  def get_gist(self, q,answers):
+    ''' extract short answer from BERT
+        using query and a few dozen doctalk best answers
+    '''
+    if self.params.with_bert_qa ==0 : return
+    from transformers import pipeline
+    #ranks=[a[2] for a in answers]
+    #assert ranks==sorted(ranks,reverse=True)
+    ws=[" ".join(a[1]) for a in answers]
+    lens=[len(a[1]) for a in answers]
+    token_count = sum(lens)
+
+    txt=" ".join(ws)
+    r=ask_bert(txt,q)
+
+    print('\n==============>BERT SHORT ANSWER:\n',
+          'sentences: ',len(lens),
+          #'sentence lengths:',lens,
+          'tokens:',token_count,':','\n',r+'\n')
+
+
+  def distill(self,q,answers,answerer):
+    '''
+    overridable answer distillation opertation
+    '''
+    self.get_gist(q,answers)
+
 
   def say(self,what):
     ''' prints and ptionally says it, unless set to quiet'''
@@ -798,7 +1021,7 @@ class Talker :
   def show_summary(self):
     ''' prints/says summary'''
     self.say('SUMMARY:')
-    for r,x,ws in take(self.params.top_sum,self.summary):
+    for r,x,ws in self.get_summary() :
       if not self.params.with_refiner :
         print(x,end=': ')
       self.say(nice(ws))
@@ -807,19 +1030,25 @@ class Talker :
   def show_keywords(self):
     ''' prints keywords'''
     print('KEYWORDS:')
-    for w in take(self.params.top_keys,nice_keys(self.keywords)) :
+    for w in self.get_keys():
       print(w)
     print('')
 
 
   def save_summary(self,out_file):
+    '''
+    saves summary as plain text for ROUGE evaluation
+    '''
     with open(out_file,'w') as g:
-      for _, _, ws in take(self.params.top_sum, self.summary):
+      for _, _, ws in self.get_summary():
         print(nice(ws),file=g)
 
   def save_keywords(self,out_file):
+    '''
+      saves keyphrases one per line for ROUGE evaluation
+    '''
     with open(out_file,'w') as g:
-      for w in take(self.params.top_keys,nice_keys(self.keywords)) :
+      for w in self.get_keys() :
         print(w,file=g)
 
   def show_rels(self):
@@ -829,10 +1058,13 @@ class Talker :
        print(svoi)
 
   def show_svos(self):
+    '''
+    optionally shows highest ranked SVO relations as graph
+    '''
     show = self.params.show_pics
     g = self.to_svo_graph()
     seeds = take(self.params.subgraph_size,
-          [x for x, r in rank_sort(self.pr) if isinstance(x, str)])
+                 (x for x, r in self.by_rank if isinstance(x, str)))
     g = g.subgraph(seeds)
     self.show_svo_graph(g,file_name=self.from_file)
 
@@ -876,15 +1108,22 @@ class Talker :
 # helpers
 
 def nice_keys(keywords):
+    '''
+      joins coumpound keyphrases, if needed
+    '''
     for w in keywords:
       if isinstance(w,tuple) :
         yield " ".join(w)
       else :
         yield w
 
-def is_clean_sent(ls) :
-  goods=[w for w in ls if w.isalpha() and w in words]
-  return len(goods)>0.8*len(ls)
+def is_clean_sent(ls,known_ratio) :
+  '''
+  heuristic on ensuring extracted sentences
+  contain mostly known English words
+  '''
+  goods=[w for w in ls if w.isalpha() and w in wnet_words]
+  return len(goods)>known_ratio*len(ls)
 
 def nice(ws) :
   ''' aggregates word lists into a nicer looking sentence'''
@@ -928,11 +1167,13 @@ def good_tag(tag,starts="NVJA"):
 
 def distinct(g) :
   '''ensures repetititions are removed from a generator'''
-  seen=set()
-  for x in g :
-    if not x in seen :
-      seen.add(x)
-      yield x
+  yield from OrderedDict.fromkeys(g)
+
+def remdup(seq) :
+  '''
+    removes duplicates in sequence in O(N) time
+  '''
+  return list(OrderedDict.fromkeys(seq))
 
 def take(k,g) :
   ''' generates only the first k elements of a sequence'''
@@ -940,20 +1181,33 @@ def take(k,g) :
     if i>=k : break
     yield x
 
-# pdf to txt conversion with external tool - optional
+
 
 def pdf2txt(fname) :
+  '''
+    pdf to txt conversion with external tool - optional
+  '''
   subprocess.run(["pdftotext", fname])
 
-# extracts file name from path
+
 def path2fname(path) :
+  '''
+    extracts file name from path
+  '''
   return path.split('/')[-1]
 
-# trimms suffix
+
 def trimSuf(path) :
+  '''
+    trimms suffix of in path+file
+  '''
   return ''.join(path.split('.')[:-1])
 
 def justFname(path) :
+  '''
+     returns just the name of the file
+     no directory path, no suffix
+  '''
   return trimSuf(path2fname(path))
 
 
